import contextlib
import torch
import torch.nn as nn
import torch.nn.functional as F


@contextlib.contextmanager
def _disable_tracking_bn_stats(model):

    def switch_attr(m):
        if hasattr(m, 'track_running_stats'):
            m.track_running_stats ^= True

    model.apply(switch_attr)
    yield
    model.apply(switch_attr)


def _l2_normalize(d):
    d_reshaped = d.view(d.shape[0], -1, *(1 for _ in range(d.dim() - 2)))
    d /= torch.norm(d_reshaped, dim=1, keepdim=True) + 1e-8
    return d




def compute_cond_entropy(logits, y=None, reduce=True):
  pyx = F.softmax(logits, dim=1)
  log_pyx = logits - torch.logsumexp(logits, dim=1, keepdim=True)
  cond_ents = -1 * (pyx * log_pyx).sum(dim=1)
  if reduce:
    return cond_ents.mean(dim=0)
  return cond_ents

def compute_entropy(logits, y=None, reduce=True):
  N = logits.size(0)
  logN = math.log(N)
  centered_logits = logits - torch.logsumexp(logits, dim=1, keepdim=True)
  log_py =torch.logsumexp(centered_logits, dim=0) - logN  # K
  py = log_py.exp()
  HY = -1. * (py * log_py).sum()
  return HY


def compute_mi(logits, y=None, reduce=True):
  # I(X, Y) = H[Y] - H[Y|X]
  HYX = compute_cond_entropy(logits, y=y, reduce=reduce)
  # p(Y) = E[P(Y|X)] = \sum P(Y|x)
  # marginalizing from logits
  N = logits.size(0)
  logN = math.log(N)
  centered_logits = logits - torch.logsumexp(logits, dim=1, keepdim=True)
  log_py =torch.logsumexp(centered_logits, dim=0) - logN  # K
  py = log_py.exp()
  HY = -1. * (py * log_py).sum()
  return HY - HYX

class VATLoss(nn.Module):

    def __init__(self, xi=10.0, eps=1.0, ip=1):
        """VAT loss
        :param xi: hyperparameter of VAT (default: 10.0)
        :param eps: hyperparameter of VAT (default: 1.0)
        :param ip: iteration times of computing adv noise (default: 1)
        """
        super(VATLoss, self).__init__()
        self.xi = xi
        self.eps = eps
        self.ip = ip

    def forward(self, model, x):
        with torch.no_grad():
            pred = F.softmax(model(x), dim=1)

        # prepare random unit tensor
        d = torch.rand(x.shape).sub(0.5).to(x.device)
        d = _l2_normalize(d)

        with _disable_tracking_bn_stats(model):
            # calc adversarial direction
            for _ in range(self.ip):
                d.requires_grad_()
                pred_hat = model(x + self.xi * d)
                # logp_hat = F.log_softmax(pred_hat, dim=1)
                # adv_distance = F.kl_div(logp_hat, pred, reduction='batchmean')
                adv_distance = compute_cond_entropy(pred_hat)
                adv_distance.backward()
                d = _l2_normalize(d.grad)
                model.zero_grad()

            # calc LDS
            r_adv = d * self.eps
            pred_hat = model(x + r_adv)
            # logp_hat = F.log_softmax(pred_hat, dim=1)
            # lds = F.kl_div(logp_hat, pred, reduction='batchmean')
            lds = compute_cond_entropy(pred_hat)

        return -1 * lds

if __name__ == '__main__':
  pass
